import asyncio
import logging
import time
from asyncio import Future
from typing import (
    Any,
    Awaitable,
    Callable,
    Concatenate,
    Iterable,
    Literal,
    ParamSpec,
    TypeVar,
)

import aiohttp
from pydantic import Field, HttpUrl, validate_call
from pymongo.errors import DuplicateKeyError

import ton_connect.model.app.response as app_responses
import ton_connect.model.wallet.event as wallet_events
from ton_connect.bridge import Bridge, BridgeMessage, Connection, Session
from ton_connect.misc import SSL_CONTEXT
from ton_connect.model.app.request import (
    AppRequestType,
    ConnectRequest,
    SendTransactionRequest, TonAddressRequestItem,
    TonProofRequestItem,
)
from ton_connect.model.app.wallet import WalletApp
from ton_connect.model.model import BaseModel
from ton_connect.model.wallet.device import Device
from ton_connect.model.wallet.event import (
    TonAddressItem,
    WalletEventName,
    WalletEventType,
)
from ton_connect.model.wallet.wallet import Account
from ton_connect.storage import BridgeStorage, BridgeData, BridgeKey

T = TypeVar("T")
C = TypeVar("C")
TC = TypeVar("TC", bound="TonConnect")
P = ParamSpec("P")

Decorator = Callable[
    Concatenate[TC, P],
    Callable[Concatenate[TC, P], Callable[Concatenate[TC, P], Awaitable[C]]],
]
Decorated = Callable[Concatenate[TC, P], Awaitable[C]]

D = Callable[[Concatenate[TC, P], Awaitable[None]], None]

LOG = logging.getLogger(__name__)

EventListener = Callable[["ConnectorEvent"], Awaitable[None]]

P = ParamSpec("P")
R = TypeVar("R")


ListenerEvent = WalletEventName | Literal["heartbeat", "stopped"]


class ConnectorEvent(BaseModel):
    wallet_name: str = Field(..., description="Wallet name")
    event: WalletEventType = Field(..., description="Event")
    device: Device = Field(..., description="User device info")
    account: Account = Field(..., description="User account info")
    entity_id: str = Field(..., description="Entity ID")


class Task:
    def __init__(self, func: Callable[P, R], *args: P.args) -> None:
        self.func = func
        self.args = args

    async def __call__(self) -> R:
        try:
            result = self.func(*self.args)
            if asyncio.iscoroutine(result):
                return await result
            return result
        except Exception as e:
            LOG.error(f"Error processing task {self.func}: {e}")


class ConnectionExistsError(Exception):
    pass


class RPCError(Exception):
    pass


class ConnectionNotFoundError(Exception):
    pass


class ConnectionSourceNotFoundError(Exception):
    pass


class ConnectionSessionNotFoundError(Exception):
    pass


class TonConnect:
    APPS = {}
    APPS_CACHE_TTL = 10 * 60
    APPS_URL = "https://raw.githubusercontent.com/ton-blockchain/wallets-list/main/wallets-v2.json"

    def __init__(
        self,
        manifest_url: HttpUrl,
        storage: BridgeStorage,
    ) -> None:
        """Init TON Connector.

        :param manifest_url: URL to manifest file of your app.
        :param storage: Storage for connection data.
        """

        self.manifest_url: HttpUrl = manifest_url
        self.storage: BridgeStorage = storage

        self.queue: asyncio.Queue[BridgeMessage] = asyncio.Queue()
        self.bridges: dict[str, Bridge] = {}
        self.lock = asyncio.Lock()
        self.send_lock = asyncio.Lock()

        self.listeners: dict[ListenerEvent, EventListener] = {}

        self.listener_started = asyncio.Event()
        self.listener: asyncio.Task | None = None

        self.rpc_response_waiters: dict[int, asyncio.Future[Any]] = {}

    def set_bridge(self, app_name: str, bridge: Bridge) -> None:
        self.bridges[app_name] = bridge

    def get_bridge(self, app_name: str) -> Bridge | None:
        return self.bridges.get(app_name)

    @staticmethod
    def ensure_listener(func: Decorator) -> Decorated:
        async def wrapper(self: TC, *args: P.args, **kwargs: P.kwargs) -> C:
            async with self.lock:
                if self.listener is None:
                    self.listener = asyncio.create_task(self.start_listener())
                    await self.listener_started.wait()
            return await func(self, *args, **kwargs)

        return wrapper

    @classmethod
    async def get_wallets(
        cls,
        app_names: list[str] | None = None,
        names: list[str] | None = None,
        ton_dns: list[str] | None = None,
        only_supported: bool = True,
        platforms: list[str] | None = None,
    ) -> list[WalletApp]:
        """Get list of supported wallet apps.

        :param app_names: List of wallet app names to filter wallets.
        :param names: List of wallet names to filter wallets.
        :param ton_dns: List of TON DNS names to filter wallets.
        :param only_supported: Get only supported wallets (python compatible).
        :param platforms: List of platforms to filter wallets.
        :return: List of wallet apps.
        """

        if cls.APPS.get("last_timestamp", 0) + cls.APPS_CACHE_TTL < time.time():
            async with aiohttp.ClientSession() as session:
                async with session.get(cls.APPS_URL, ssl=SSL_CONTEXT) as response:
                    response_apps = [
                        WalletApp.model_validate(wallet)
                        for wallet in await response.json(content_type="text/plain")
                    ]
                    cls.APPS["last_timestamp"] = time.time()
                    cls.APPS["apps"] = response_apps

        apps: Iterable[WalletApp] = (app for app in cls.APPS["apps"])

        apps = filter(lambda app: app.name in names, apps) if names else apps
        apps = (
            filter(lambda app: set(app.platforms).intersection(platforms), apps)
            if platforms
            else apps
        )
        apps = filter(lambda app: app.is_supported, apps) if only_supported else apps
        apps = filter(lambda app: app.app_name in app_names, apps) if app_names else apps
        apps = filter(lambda app: app.dns in ton_dns, apps) if ton_dns else apps

        return list(apps)

    @ensure_listener
    @validate_call
    async def connect(self, wallet: WalletApp, ton_proof: TonProofRequestItem | None = None) -> str:
        """Connect to the wallet.

        :param wallet: Wallet to connect.
        :param ton_proof: TON proof request item.
        :return: Connection URL.
        """

        async with self.lock:
            connection = await self.storage.get_connection(wallet.app_name)
            if connection is not None and connection.connect_event:
                raise ConnectionExistsError(
                    "Connection already exists. Use restore_connection method."
                )

            bridge = self.get_bridge(wallet.app_name)
            if bridge is not None and bridge.is_alive:
                await bridge.disconnect(send_event=True)
                await asyncio.sleep(0.2)

            await self.storage.delete(wallet.app_name)
            await self.storage.insert(wallet.app_name, BridgeData())

            ready = asyncio.Event()

            bridge = Bridge(
                wallet.app_name,
                self.queue,
                connector_ready=ready,
                bridge_url=wallet.bridge_url,
                universal_url=wallet.universal_url,
            )

            await bridge.register_session()
            await bridge.connected.wait()

            self.set_bridge(wallet.app_name, bridge)

            if connection is None:
                session = Session(
                    private_key=bridge.crypto.private_key.encode().hex(),
                    bridge_url=bridge.bridge_url,
                )

                connection = Connection(
                    session=session,
                    source=wallet.app_name,
                )

            await self.storage.set_connection(wallet.app_name, connection)

            request_items: list[TonAddressRequestItem | TonProofRequestItem] = [
                TonAddressRequestItem()
            ]
            if ton_proof:
                request_items.append(ton_proof)

            request = ConnectRequest(manifest_url=str(self.manifest_url), items=request_items)
            ready.set()

            return bridge.generate_connect_url(request, bridge.crypto.public_key)

    @ensure_listener
    @validate_call
    async def restore_connection(self, wallet: WalletApp) -> Bridge:
        """Restore connection to the wallet."""

        async with self.lock:
            connection = await self.storage.get_connection(wallet.app_name)
            if not connection:
                LOG.info("Connection not found for %s. Use .connect", wallet.app_name)
                raise ConnectionNotFoundError()

            if not connection.source:
                LOG.info("Connection source not found for %s", wallet.app_name)
                raise ConnectionSourceNotFoundError()

            if not connection.session:
                LOG.info("Connection session not found for %s", wallet.app_name)
                raise ConnectionSessionNotFoundError()

            ready = asyncio.Event()

            bridge = Bridge(
                wallet.app_name,
                self.queue,
                connector_ready=ready,
                bridge_url=connection.session.bridge_url,
                universal_url=connection.session.bridge_url,
                private_key=connection.session.private_key.hex(),
                last_rpc_event_id=connection.last_rpc_event_id,
            )

            await bridge.register_session()

            self.set_bridge(wallet.app_name, bridge)

            ready.set()

            LOG.debug(
                "Waiting for bridge connection %s %s",
                self.storage.entity_id,
                wallet.app_name,
            )

            await bridge.connected.wait()

            LOG.debug("Bridge connected %s %s", self.storage.entity_id, wallet.app_name)

            return bridge

    async def handle_message(self, connection: Connection, message: BridgeMessage) -> None:
        """Handle queue message."""

        int_tasks: list[Task] = []

        if message.event != "heartbeat":
            LOG.info("Handling message: %s", message)

        match message.event:
            case "heartbeat":
                LOG.debug("Heartbeat received")
                await self.storage.set(message.app_name, BridgeKey.HEARTBEAT, int(time.time()))

            case "stopped":
                LOG.info("Bridge %s stopped", message.app_name)
                self.bridges.pop(message.app_name)
                await self.storage.remove(message.app_name, BridgeKey.CONNECTION)

            case wallet_events.ConnectSuccessEvent():
                connection.last_wallet_event_id = message.event.id
                if message.event.payload.find_item_by_type(TonAddressItem) is not None:
                    connection.session.wallet_key = message.source
                    connection.connect_event = message.event
                await self.storage.set_connection(message.app_name, connection)

            case wallet_events.DisconnectEvent() | wallet_events.ConnectErrorEvent():
                LOG.info("Disconnecting from %s for %s", message.app_name, self.storage.entity_id)

                bridge = self.get_bridge(message.app_name)
                int_tasks.append(Task(bridge.disconnect))
                int_tasks.append(Task(self.storage.delete, message.app_name))

            case (
                app_responses.SendTransactionResponseError()
                | app_responses.SendTransactionSuccess()
                | app_responses.SignDataResponseError()
                | app_responses.SignDataSuccess()
            ):
                if message.event.id in self.rpc_response_waiters:
                    self.rpc_response_waiters[message.event.id].set_result(message)
                else:
                    LOG.error("Unexpected App message: %s", message)

                connection.last_rpc_event_id = message.event.id
                return

            case _:
                LOG.error(f"Unhandled event: {message.event}")

        if isinstance(message.event, WalletEventType) and message.event.name in self.listeners:
            connection = await self.storage.get_connection(message.app_name)
            connector_event = ConnectorEvent(
                wallet_name=message.app_name,
                event=message.event,
                device=connection.connect_event.payload.device,
                account=connection.connect_event.payload.find_item_by_type(TonAddressItem),
                entity_id=self.storage.entity_id,
            )
            await asyncio.create_task(self.listeners[message.event.name](connector_event))
        elif isinstance(message.event, WalletEventType):
            LOG.error(f"Unhandled event: {message.event}")

        for task in int_tasks:
            await task()

    async def start_listener(self) -> None:
        """Listen for wallet events."""

        LOG.debug("Starting TonConnector event listener...")

        self.listener_started.set()

        try:
            while True:
                try:
                    message = await self.queue.get()
                    LOG.debug(f"Event received: {message}")

                    async with self.lock:
                        connection = await self.storage.get_connection(message.app_name)
                        if connection is None:
                            LOG.error(f"Connection not found for {message.app_name}")
                            continue

                        await asyncio.wait_for(
                            self.handle_message(connection, message),
                            timeout=5,
                        )

                except Exception as e:
                    LOG.error(f"Error processing event: {e}")
                else:
                    self.queue.task_done()

        except asyncio.CancelledError:
            LOG.debug("TonConnector event listener stopped. Stopping bridge listeners...")
            for bridge in self.bridges.values():
                await bridge.disconnect(send_event=False)
        finally:
            self.listener = None
            self.listener_started.clear()

    async def stop_listener(self) -> None:
        """Stop listener."""

        if self.listener is not None:
            self.listener.cancel()
            await self.listener

    @ensure_listener
    @validate_call
    async def listen(
        self,
        event: ListenerEvent,
        handler: EventListener,
    ) -> None:
        """Add listener to the event.

        :param event: Event name.
        :param handler: Event handler.
        """

        if event in self.listeners:
            raise ValueError(f"Event {event} is already registered")

        self.listeners[event] = handler

    @validate_call
    async def send(
        self,
        app_name: str,
        request: AppRequestType,
        wait_response: bool = True,
        timeout: int = 5,
    ) -> asyncio.Task[app_responses.AppResponses] | None:
        """Send request to the wallet.

        :param app_name: Wallet app name.
        :param request: Request to send.
        :param wait_response: Wait for response. Return task if True.
        :param timeout: Timeout for sending request.
        """

        async with self.send_lock:
            bridge = self.get_bridge(app_name)
            if bridge is None:
                raise Exception("Bridge not found")

            connection = await self.storage.get_connection(app_name)
            if connection is None:
                raise RuntimeError("Connection not found")

            request.id = connection.next_rpc_request_id
            connection.next_rpc_request_id += 1

            await self.storage.set_connection(app_name, connection)

            ttl = 5 * 60

            response = await bridge.send_request(
                request,
                wallet_app_key=connection.session.wallet_key,
                ttl=ttl,
                timeout=timeout,
            )

        LOG.info("Got response for request %s: %s", request.id, response)

        if response["statusCode"] == 200 and wait_response:
            ready: asyncio.Future[app_responses.AppResponses] = Future()
            self.rpc_response_waiters[request.id] = ready

            return asyncio.create_task(asyncio.wait_for(ready, timeout=ttl))

        elif response["statusCode"] != 200:
            raise RPCError(response)

        return None



from enum import Enum
from typing import Any, Generic, List, Optional, TypeVar, Union

from pydantic import Field
from typing_extensions import Literal

from ton_connect.model.error import Error
from ton_connect.model.misc import GenericIntEnum, GenericLiteral
from ton_connect.model.model import BaseModel
from ton_connect.model.wallet.device import Device
from ton_connect.model.wallet.error import ConnectErrorCodes, ConnectItemErrorCodes
from ton_connect.model.wallet.wallet import Account, TonProof

_ReplyItemGeneric = TypeVar("_ReplyItemGeneric", bound="ReplyItem")
_IntEnumGeneric = TypeVar("_IntEnumGeneric", bound=GenericIntEnum)
_MessageType = TypeVar("_MessageType")
_LiteralGeneric = TypeVar("_LiteralGeneric")
_PayloadGeneric = TypeVar("_PayloadGeneric", bound="Payload")


class ConnectReplyItems(str, Enum):
    TON_ADDR = "ton_addr"
    TON_PROOF = "ton_proof"


class WalletEventName(str, Enum):
    CONNECT = "connect"
    CONNECT_ERROR = "connect_error"
    DISCONNECT = "disconnect"


class ReplyItem(BaseModel, Generic[_LiteralGeneric]):
    """Base class for reply items."""

    name: GenericLiteral[_LiteralGeneric] = Field(..., description="Name of the item")


class Payload(BaseModel, Generic[_ReplyItemGeneric]):
    items: List[_ReplyItemGeneric] = Field(..., description="List of reply items")

    def find_item_by_type(self, item_type: type[_ReplyItemGeneric]) -> Optional[_ReplyItemGeneric]:
        """Find item by type."""

        for item in self.items:
            if isinstance(item, item_type):
                return item

        return None


class Event(BaseModel, Generic[_LiteralGeneric, _PayloadGeneric]):
    name: _LiteralGeneric = Field(..., description="Event name", alias="event")
    id: int = Field(..., description="Event ID")
    payload: _PayloadGeneric = Field(..., description="Event payload")


ConnectItemError = Error[ConnectItemErrorCodes, Optional[str]]
ConnectEventError = Error[ConnectErrorCodes, str]


class ItemErrorReply(BaseModel):
    error: ConnectItemError = Field(..., description="Error item")


class TonAddressItem(ReplyItem[Literal[ConnectReplyItems.TON_ADDR]], Account):
    pass


class TonProofSuccessItem(ReplyItem[Literal[ConnectReplyItems.TON_PROOF]]):
    proof: TonProof = Field(..., description="Proof of the user's account")


class TonProofFailureItem(ReplyItem[Literal[ConnectReplyItems.TON_PROOF]], ItemErrorReply):
    pass


TonProofItemsType = Union[TonProofSuccessItem, TonProofFailureItem]
ConnectItemsType = Union[TonAddressItem, TonProofItemsType]


class ConnectSuccessPayload(Payload[ConnectItemsType]):
    device: Device = Field(..., description="User's device info")


ConnectSuccessEvent = Event[Literal[WalletEventName.CONNECT], ConnectSuccessPayload]
ConnectErrorEvent = Event[Literal[WalletEventName.CONNECT_ERROR], ConnectEventError]
DisconnectEvent = Event[Literal[WalletEventName.DISCONNECT], dict[str, Any]]

WalletEventType = ConnectSuccessEvent | ConnectErrorEvent | DisconnectEvent
ConnectEventType = ConnectSuccessEvent | ConnectErrorEvent



import asyncio
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Generic, TypeVar, cast

from pydantic import BaseModel, ConfigDict, Field
from pymongo.asynchronous.collection import AsyncCollection
from pymongo.asynchronous.mongo_client import AsyncMongoClient

from ton_connect.bridge import Connection

Model = TypeVar("Model", bound="StorageData")


class BridgeKey(str, Enum):
    CONNECTION = "connection"
    LAST_EVENT_ID = "last_event_id"
    HEARTBEAT = "heartbeat"


class BridgeData(BaseModel):
    model_config = ConfigDict(extra="allow")

    connection: Connection | None = Field(None, description="Connection to wallet object.")
    last_event_id: int | None = Field(None, description="Last wallet event ID.")


class BridgeStorage(ABC, Generic[Model]):
    def __init__(self, entity_id: str) -> None:
        self.entity_id = entity_id
        self.lock = asyncio.Lock()

    @abstractmethod
    async def insert(self, app_name: str, data: Model) -> None: ...

    @abstractmethod
    async def set(self, app_name: str, key: BridgeKey, value: Connection | int) -> None: ...

    @abstractmethod
    async def get(self, app_name: str, key: BridgeKey) -> Connection | int: ...

    @abstractmethod
    async def remove(self, app_name: str, key: BridgeKey) -> None: ...

    @abstractmethod
    async def delete(self, app_name: str) -> None: ...

    async def get_connection(self, app_name: str) -> Connection | None:
        """Get connection from storage."""

        async with self.lock:
            try:
                return cast(Connection, await self.get(app_name, BridgeKey.CONNECTION))
            except KeyError:
                return None

    async def set_connection(self, app_name: str, connection: Connection) -> None:
        """Set connection to storage."""

        async with self.lock:
            await self.set(app_name, BridgeKey.CONNECTION, connection)

    async def get_last_event_id(self, app_name: str) -> int | None:
        async with self.lock:
            return cast(int | None, await self.get(app_name, BridgeKey.LAST_EVENT_ID))

    async def set_last_event_id(self, app_name: str, value: int) -> None:
        async with self.lock:
            await self.set(app_name, BridgeKey.LAST_EVENT_ID, value)


class DictBridgeStorage(BridgeStorage, Generic[Model]):
    STORAGE: dict[str, Model] = {}

    @staticmethod
    def gen_key(app_name: str, key: str) -> str:
        return f"{app_name}:{key}"

    async def delete(self, app_name: str):
        """Delete storage."""

        key = self.gen_key(app_name, self.entity_id)
        if key in self.STORAGE:
            del self.STORAGE[key]


    async def insert(self, app_name: str, data: Model) -> None:
        """Insert user wallet app to storage."""

        key = self.gen_key(app_name, self.entity_id)
        if key in self.STORAGE:
            raise KeyError("App already exists in storage.")

        self.STORAGE[key] = data

    async def set(self, app_name: str, key: BridgeKey, value: Connection | int) -> None:
        setattr(self.STORAGE[self.gen_key(app_name, self.entity_id)], key.value, value)

    async def get(
        self,
        app_name: str,
        key: BridgeKey,
    ) -> Connection | int:
        return getattr(self.STORAGE[self.gen_key(app_name, self.entity_id)], key)

    async def remove(self, app_name: str, key: BridgeKey) -> None:
        delattr(self.STORAGE[self.gen_key(app_name, self.entity_id)], key.value)


class MongoBridgeStorage(BridgeStorage, Generic[Model]):
    def __init__(
        self,
        entity_id: str,
        client: AsyncMongoClient[dict[str, Any]],
        database: str,
        collection: str,
        primary_key: str = "_id",
    ) -> None:
        super().__init__(entity_id)

        self.client = client
        self.database_name = database
        self.collection_name = collection
        self.primary_key = primary_key

    def gen_search_query(self, app_name: str) -> dict[str, Any]:
        """Generate search query for storage."""

        return {
            self.primary_key: self.entity_id,
            "app_name": app_name,
        }

    @property
    def collection(self) -> AsyncCollection[dict[str, Any]]:
        return self.client[self.database_name][self.collection_name]

    async def ensure_index(self) -> None:
        """Ensure index for storage."""

        await self.collection.create_index([(self.primary_key, 1), ("app_name", 1)], unique=True)
        await self.collection.create_index("heartbeat")

    async def delete(self, app_name: str) -> None:
        """Delete app storage."""

        await self.collection.delete_one(self.gen_search_query(app_name))

    async def insert(self, app_name: str, data: Model) -> None:
        """Insert user wallet app to storage."""

        await self.collection.insert_one(
            {
                self.primary_key: self.entity_id,
                "app_name": app_name,
                "wallet": data.model_dump(
                    mode="json",
                    by_alias=True,
                    exclude_none=True,
                ),
            }
        )

    async def get(
        self,
        app_name: str,
        key: BridgeKey,
    ) -> Connection | int:
        """Get value from storage. Requires extra motor dependency."""

        projection = {key.value: 1, self.primary_key: 1}

        doc = await self.collection.find_one(self.gen_search_query(app_name), projection)
        if not doc:
            raise KeyError("App not found in storage.")

        return getattr(BridgeData.model_validate(doc), key.value)

    async def set(self, app_name: str, key: BridgeKey, value: Connection | int) -> None:
        """Set value to storage. Requires extra motor dependency."""

        await self.collection.update_one(
            self.gen_search_query(app_name),
            {
                "$set": {
                    key.value: value
                    if isinstance(value, int)
                    else value.model_dump(mode="json", by_alias=True, exclude_none=True)
                }
            },
        )

    async def remove(self, app_name: str, key: BridgeKey) -> None:
        """Remove value from storage. Requires extra motor dependency."""

        await self.collection.update_one(
            self.gen_search_query(app_name),
            {"$unset": {key.value: ""}},
        )



from typing import Generic, TypeVar

from pydantic import Field

from ton_connect.model.error import Error, _IntEnumGeneric, _MessageType
from ton_connect.model.model import BaseModel

_GenericResponseError = TypeVar("_GenericResponseError", bound="ResponseError")
_DataType = TypeVar("_DataType")
_ResultGeneric = TypeVar("_ResultGeneric")


class ResponseSuccess(BaseModel, Generic[_ResultGeneric]):
    id: int = Field(..., description="Event ID")
    result: _ResultGeneric = Field(..., description="Result of the action")


class ResponseError(
    Error[_IntEnumGeneric, _MessageType],
    Generic[_IntEnumGeneric, _MessageType, _DataType],
):
    data: _DataType = Field(None, description="Error data")



from datetime import datetime, timedelta
from enum import Enum
from typing import Annotated, Any, ClassVar, Generic, List, Literal, TypeVar

from pydantic import Field, computed_field
from pydantic.main import IncEx

from ton_connect.model.model import BaseModel

_Method = TypeVar("_Method")


class Method(str, Enum):
    SEND_TRANSACTION = "sendTransaction"
    SIGN_DATA = "signData"
    DISCONNECT = "disconnect"


class AppRequest(BaseModel, Generic[_Method]):
    """Base class for request items."""

    PARAMETERS: ClassVar[list[str]] = []

    method: _Method = Field(..., description="Method name")
    id: int | None = Field(None, description="Event ID")

    @computed_field
    def params(self) -> List[str]:
        return [
            super().model_dump_json(
                include={param for param in self.PARAMETERS},
                by_alias=True,
                exclude_none=True,
            )
        ]

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: dict[str, Any] | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        return super().model_dump(
            mode=mode,
            include={"id", "method", "params"},
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
            exclude_unset=exclude_unset,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=serialize_as_any,
        )


class SignDataParams(BaseModel):
    schema_crc: int = Field(..., description="Schema version", alias="schema_crc")
    cell: str = Field(..., description="Cell to sign")


class SendTransactionMessage(BaseModel):
    address: str = Field(..., description="Address")
    amount: str = Field(..., description="Amount")


class SendTransactionRequest(AppRequest[Literal[Method.SEND_TRANSACTION]]):
    PARAMETERS = ["valid_until", "address", "network", "messages"]

    method: Literal[Method.SEND_TRANSACTION] = Method.SEND_TRANSACTION
    valid_until: int = Field(
        default_factory=lambda: int((datetime.now() + timedelta(minutes=30)).timestamp()),
        description="Request expiration time",
    )
    address: str = Field(..., description="Address", alias="from")
    network: int = Field(..., description="Network id")
    messages: List[SendTransactionMessage] = Field(..., description="List of messages")


class SignDataRequest(AppRequest[Literal[Method.SIGN_DATA]]):
    method: Literal[Method.SIGN_DATA] = Method.SIGN_DATA


class DisconnectRequest(AppRequest[Literal[Method.DISCONNECT]]):
    method: Literal[Method.DISCONNECT] = Method.DISCONNECT


class TonAddressRequestItem(BaseModel):
    name: Literal["ton_addr"] = Field("ton_addr", description="Name of the item", init=False)


class TonProofRequestItem(BaseModel):
    name: Literal["ton_proof"] = Field(
        "ton_proof",
        description="Name of the item",
        init=False,
    )
    payload: str = Field(..., description="TON proof payload")


class ConnectRequest(BaseModel):
    method: ClassVar[str] = "connect"

    manifest_url: Annotated[
        str,
        Field(..., description="URL to manifest", alias="manifestUrl"),
    ]
    items: list[TonAddressRequestItem | TonProofRequestItem] = Field(
        default_factory=list,
        description="List of items",
    )


AppRequestType = DisconnectRequest | SendTransactionRequest | SignDataRequest



import pydantic


class BaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        populate_by_name=True,
        json_encoders={
            bytes: lambda v: v.hex(),
        },
    )



import base64
import json
from typing import Any

from nacl.encoding import HexEncoder
from nacl.public import Box, PrivateKey, PublicKey
from nacl.utils import random


class SessionCrypto:
    def __init__(self, private_key: PrivateKey | str | bytes | None = None) -> None:
        """Session crypto class."""

        if isinstance(private_key, (str, bytes)):
            private_key = private_key.encode() if isinstance(private_key, str) else private_key
            private_key = PrivateKey(private_key, HexEncoder)

        elif private_key is None:
            private_key = PrivateKey.generate()

        self.private_key: PrivateKey = private_key
        self.public_key: str = self.private_key.public_key.encode().hex()

    @staticmethod
    def generate_nonce() -> bytes:
        return random(Box.NONCE_SIZE)

    def encrypt(self, data: dict[str, Any], pub_key: PublicKey | str | bytes) -> bytes:
        """Encrypt data with public private_key

        :param data: Data to encrypt.
        :param pub_key: Recipient public private_key.
        :return: Encrypted data.
        """

        if isinstance(pub_key, (str, bytes)):
            pub_key = pub_key.encode() if isinstance(pub_key, str) else pub_key.hex().encode()
            pub_key = PublicKey(pub_key, HexEncoder)

        box = Box(self.private_key, pub_key)
        nonce = self.generate_nonce()
        encrypted_data = box.encrypt(json.dumps(data).encode(), nonce)

        return base64.b64encode(nonce + encrypted_data.ciphertext)

    def decrypt(self, data: bytes, pub_key: PublicKey | str | bytes) -> str:
        """Decrypt data with public private_key

        :param data: Data to decrypt.
        :param pub_key: Sender public private_key.
        :return: Decrypted data.
        """

        if isinstance(pub_key, (str, bytes)):
            pub_key = pub_key.encode() if isinstance(pub_key, str) else pub_key.hex().encode()
            pub_key = PublicKey(pub_key, HexEncoder)

        box = Box(self.private_key, pub_key)
        nonce = data[: Box.NONCE_SIZE]
        encrypted_data = data[Box.NONCE_SIZE :]

        return box.decrypt(encrypted_data, nonce).decode()



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field, field_validator
from pydantic_core.core_schema import ValidationInfo


class BridgeType(str, Enum):
    SSE = "sse"
    JS = "js"


class WalletBridge(BaseModel):
    type: BridgeType = Field(description="Type of the bridge")
    url: Optional[str] = Field(None, description="URL to the bridge. Present if type is SSE")
    key: Optional[str] = Field(None, description="Key to the bridge. Present if type is JS")

    @field_validator("url", mode="before")
    @classmethod
    def validate_sse_url(cls, v: str, validation_info: ValidationInfo) -> str:
        if validation_info.data["type"] == BridgeType.SSE and not v:
            raise ValueError("URL is required for SSE bridge")
        return v

    @field_validator("key", mode="before")
    @classmethod
    def validate_js_key(cls, v: str, validation_info: ValidationInfo) -> str:
        if validation_info.data["type"] == BridgeType.JS and not v:
            raise ValueError("Key is required for JS bridge")
        return v


class WalletApp(BaseModel):
    app_name: str = Field(description="Internal name of the wallet")
    name: str = Field(description="Name of the wallet")
    image: str = Field(description="Image of the wallet logo")
    dns: Optional[str] = Field(None, description="TON DNS of the wallet", alias="tondns")
    about_url: str = Field(description="URL to the wallet description")
    universal_url: Optional[str] = Field(None, description="URL to the wallet app")
    bridge: List[WalletBridge] = Field(description="List of bridges")
    deep_link: Optional[str] = Field(None, description="Deep link to the wallet", alias="deepLink")
    platforms: List[str] = Field(default_factory=list, description="List of supported platforms")

    @property
    def bridge_url(self) -> Optional[str]:
        """Gets SSE bridge url if available."""

        for bridge in self.bridge:
            if bridge.type == BridgeType.SSE:
                return bridge.url

        return None

    @property
    def is_supported(self) -> bool:
        """Check if wallet is supported."""

        return self.universal_url is not None or self.bridge_url is not None



from typing import Generic, TypeVar

from pydantic import Field

from ton_connect.model.misc import GenericIntEnum
from ton_connect.model.model import BaseModel

_IntEnumGeneric = TypeVar("_IntEnumGeneric", bound=GenericIntEnum)
_MessageType = TypeVar("_MessageType")


class Error(BaseModel, Generic[_IntEnumGeneric, _MessageType]):
    """Base class for error items."""

    code: _IntEnumGeneric = Field(..., description="Error code")
    message: _MessageType = Field(None, description="Error message")



from enum import IntEnum


class ConnectErrorCodes(IntEnum):
    UNKNOWN_ERROR = 0
    BAD_REQUEST_ERROR = 1
    MANIFEST_NOT_FOUND_ERROR = 2
    MANIFEST_CONTENT_ERROR = 3
    UNKNOWN_APP_ERROR = 100
    USER_REJECTS_ERROR = 300
    METHOD_NOT_SUPPORTED = 400


class ConnectItemErrorCodes(IntEnum):
    UNKNOWN_ERROR = 0
    METHOD_NOT_SUPPORTED = 400



from enum import IntEnum
from typing import Literal, TypeVar

from typing_extensions import Annotated, TypeAliasType

T = TypeVar("T")

GenericIntEnum = TypeAliasType("GenericIntEnum", Annotated[IntEnum, T], type_params=(T,))
GenericLiteral = TypeAliasType("GenericLiteral", Literal[T], type_params=(T,))  # type:ignore[valid-type]



from typing import Any, Optional

from ton_connect.model.app.error import (
    AppResponseError,
    DisconnectErrorCode,
    SignDataErrorCode,
    TransactionErrorCode,
)
from ton_connect.model.response import ResponseError, ResponseSuccess
from ton_connect.model.wallet.response import SignDataSuccessResponse

SendTransactionResponseError = AppResponseError[
    ResponseError[TransactionErrorCode, str, Optional[Any]]
]

SendTransactionSuccess = ResponseSuccess[str]
DisconnectResponseError = AppResponseError[ResponseError[DisconnectErrorCode, str, Optional[Any]]]
SignDataResponseError = AppResponseError[ResponseError[SignDataErrorCode, str, Optional[Any]]]
SignDataSuccess = ResponseSuccess[SignDataSuccessResponse]

AppResponses = (
    SendTransactionResponseError
    | SendTransactionSuccess
    | DisconnectResponseError
    | SignDataResponseError
    | SignDataSuccess
)



from typing import Annotated

from pydantic import BeforeValidator


def validated_hex_string(v: str | bytes) -> bytes:
    if isinstance(v, bytes):
        return v

    return bytes.fromhex(v)


HexBytes = Annotated[bytes, BeforeValidator(validated_hex_string)]



from enum import IntEnum
from typing import Generic, TypeVar

from pydantic import BaseModel, Field

from ton_connect.model.response import ResponseError

_GenericResponseError = TypeVar("_GenericResponseError", bound=ResponseError)


class TransactionErrorCode(IntEnum):
    UNKNOWN_ERROR = 0
    BAD_REQUEST_ERROR = 1
    UNKNOWN_APP_ERROR = 100
    USER_REJECTS_ERROR = 300
    METHOD_NOT_SUPPORTED = 400


class DisconnectErrorCode(IntEnum):
    UNKNOWN_ERROR = 0
    BAD_REQUEST_ERROR = 1
    UNKNOWN_APP_ERROR = 100
    METHOD_NOT_SUPPORTED = 400


class SignDataErrorCode(IntEnum):
    UNKNOWN_ERROR = 0
    BAD_REQUEST_ERROR = 1
    UNKNOWN_APP_ERROR = 100
    USER_REJECTS_ERROR = 300
    METHOD_NOT_SUPPORTED = 400


class AppResponseError(BaseModel, Generic[_GenericResponseError]):
    id: str = Field(..., description="Event ID")
    error: _GenericResponseError = Field(..., description="Error message")



import logging
from enum import Enum, IntEnum
from typing import Optional

from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey
from pydantic import Base64Str, BaseModel, Field

from ton_connect.types import HexBytes

LOG = logging.getLogger(__name__)


class Chain(IntEnum):
    MAINNET = "-239"
    TESTNET = "-3"


class Address(str):
    """TON wallet address."""

    @property
    def work_chain(self) -> str:
        return self.split(":", 1)[-1]

    @property
    def hash(self) -> str:
        return self.split(":", 2)[-1]


class Account(BaseModel):
    """User's wallet account."""

    address: str = Field(..., description="User's address. Format: <wc>:<hex>")
    chain: Chain = Field(..., description="Network chain", alias="network")
    wallet_state_init: str = Field(
        ...,
        description="Wallet contract state_init",
        alias="walletStateInit",
    )
    public_key: Optional[HexBytes] = Field(
        None,
        description="Public private_key of the user's account",
        alias="publicKey",
    )


class Provider(str, Enum):
    HTTP = "http"


class Wallet(BaseModel):
    """Information about user's wallet."""

    provider: Provider = Field(..., description="Type of the provider")
    account: Account = Field(..., description="User's account", alias="ton_addr")


class TonProofDomain(BaseModel):
    len: int = Field(..., description="Length of the domain", alias="lengthBytes")
    val: str = Field(..., description="Value of the domain", alias="value")


class TonProof(BaseModel):
    timestamp: int = Field(..., description="Timestamp of the proof")
    domain: TonProofDomain = Field(..., description="Domain of the proof")
    payload: str = Field(..., description="Payload of the proof")
    signature: str = Field(..., description="Signature of the proof")

    def verify(self, public_key: bytes) -> bool:
        """Verify proof signature."""

        try:
            verify_key = VerifyKey(public_key, encoder=HexEncoder)
            verify_key.verify(
                f"{self.timestamp}{self.domain.len}{self.domain.val}{self.payload}".encode(),
                self.signature.encode(),
            )
            return True
        except Exception as e:
            LOG.error("Failed to verify proof signature: %s", e)
            return False



import asyncio
import logging
import re
from datetime import timedelta
from functools import cached_property
from typing import Annotated, Any, Awaitable, Callable, Literal
from urllib.parse import quote_plus

import aiohttp
import aiohttp_sse_client.client
import pydantic
from aiohttp_sse_client.client import MessageEvent
from nacl.exceptions import CryptoError
from pydantic import AnyUrl, Base64Bytes, Field, HttpUrl, ValidationError

from ton_connect.crypto import SessionCrypto
from ton_connect.misc import SSL_CONTEXT, encode_telegram_url_parameters, iterate_event_source
from ton_connect.model.app.request import AppRequest, ConnectRequest
from ton_connect.model.app.response import AppResponses
from ton_connect.model.model import BaseModel
from ton_connect.model.wallet.event import ConnectEventType, WalletEventType
from ton_connect.types import HexBytes

LOG = logging.getLogger(__name__)


class Session(BaseModel):
    private_key: HexBytes = Field(..., description="Session private key")
    wallet_key: HexBytes | None = Field(None, description="Wallet public key")
    bridge_url: str = Field(..., description="Bridge URL")


class Connection(BaseModel):
    session: Session = Field(..., description="Session")
    source: str = Field(..., description="Source of the connection")
    next_rpc_request_id: int = Field(0, description="Next RPC request ID")
    next_event_id: int | None = Field(None, description="Next event ID")
    last_wallet_event_id: int | None = Field(None, description="Last wallet event ID")
    last_rpc_event_id: int | None = Field(None, description="Last RPC event ID")
    connect_event: ConnectEventType | None = Field(None, description="Connect event")


class WalletMessageEvent(BaseModel):
    sender: HexBytes = Field(..., description="Sender public key", alias="from")
    message: Base64Bytes = Field(..., description="Encrypted message")


class BridgeMessage(BaseModel):
    event: Literal["heartbeat", "stopped"] | WalletEventType | AppResponses = Field(
        ..., description="Wallet event"
    )
    app_name: str = Field(..., description="Wallet app name")
    source: HexBytes = Field(..., description="Source of the message")


class Bridge:
    VERSION: int = 2
    """Bridge API version."""

    UNIVERSAL_URL: str = "tc://"
    """Universal URL for the bridge."""

    TIMEOUT: int = 600
    """Timeout for the connection."""

    PATH_MESSAGE: str = "message"
    """Path to the message endpoint."""

    PATH_EVENTS: str = "events"
    """Path to the events endpoint."""

    TTL: int = 300
    """Time to live for the request."""

    @property
    def request_headers(self) -> dict[str, str]:
        """Request headers for the bridge."""

        return {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    def __init__(
        self,
        app_name: str,
        queue: asyncio.Queue[BridgeMessage],
        connector_ready: asyncio.Event,
        bridge_url: Annotated[str, HttpUrl] | None = None,
        universal_url: Annotated[str, AnyUrl] | None = None,
        private_key: str | None = None,
        last_rpc_event_id: int | None = None,
    ) -> None:
        self.app_name = app_name
        self.queue = queue
        self.bridge_url = bridge_url
        self.universal_url = universal_url or self.UNIVERSAL_URL
        self.crypto: SessionCrypto = SessionCrypto(private_key=private_key)

        self.connected: asyncio.Event = asyncio.Event()
        self.connector_ready: asyncio.Event = connector_ready
        self.listener: asyncio.Task[Callable[[str], Awaitable[None]]] | None = None

        self.last_rpc_event_id: str | None = last_rpc_event_id
        self.stop = asyncio.Event()

    @property
    def is_alive(self) -> bool:
        """Check if the bridge is alive."""

        return not self.listener.done() and not self.listener.cancelled()

    @cached_property
    def type_adapter(self) -> pydantic.TypeAdapter:
        return pydantic.TypeAdapter[AppResponses | WalletEventType](AppResponses | WalletEventType)

    def reset_crypto(self) -> None:
        """Reset crypto session."""

        self.crypto = SessionCrypto()

    @staticmethod
    def generate_basic_connect_url(
        request: ConnectRequest,
        session_id: str,
        version: int = VERSION,
        universal_url: str = UNIVERSAL_URL,
    ) -> Annotated[str, AnyUrl]:
        """Generate basic URL for the bridge."""

        params = quote_plus(request.model_dump_json(by_alias=True, exclude_none=True))

        universal_url = universal_url.rstrip("/")
        return f"{universal_url}?v={version}&id={session_id}&r={params}&ret=back"

    @staticmethod
    def convert_to_direct_link(url: str) -> str:
        """Convert universal URL to direct link."""

        if "attach=" in url:
            # remove attach query param and its value from url
            url = re.sub(r"attach=[^&]+", "", url)

            # add /start to url path
            url = re.sub(r"(\?|&)", "/start?", url)

        return url

    def generate_connect_url(
        self,
        request: ConnectRequest,
        session_id: str,
    ) -> Annotated[str, AnyUrl]:
        """Generate URL for the bridge."""

        universal_url = self.universal_url

        telegram_url = r"^(http[s]?://)?t.me/(\w+)"

        if universal_url.startswith("tg://") or re.match(telegram_url, universal_url):
            basic_url = self.generate_basic_connect_url(
                request,
                session_id,
                universal_url="",
            )

            universal_url = self.convert_to_direct_link(self.universal_url)

            params = basic_url.split("?")[1]
            start_command = f"tonconnect-{encode_telegram_url_parameters(params)}"

            if universal_url.endswith("?"):
                return f"{universal_url}startapp={start_command}"

            return f"{universal_url}&startapp={start_command}"

        return self.generate_basic_connect_url(request, session_id, universal_url=universal_url)

    async def connect(self, request: ConnectRequest) -> Annotated[str, AnyUrl]:
        """Send request to connect to the wallet via the bridge."""

        await self.disconnect(send_event=False)
        self.reset_crypto()

        return self.generate_connect_url(request, self.crypto.public_key)

    async def disconnect(self, send_event: bool = True) -> None:
        """DisconnectEvent from the wallet."""

        if self.listener is not None:
            if send_event:
                await self.queue.put(
                    BridgeMessage(
                        event="stopped",
                        app_name=self.app_name,
                        source=b"",
                    )
                )
            self.listener.cancel()

    async def send_request(
        self,
        request: AppRequest,
        wallet_app_key: HexBytes,
        ttl: int | None = None,
        timeout: int = 5,
    ) -> dict[str, Any]:
        """Send request to the wallet."""

        url = (
            f"{self.bridge_url}/{self.PATH_MESSAGE}?"
            f"client_id={self.crypto.public_key}"
            f"&to={wallet_app_key.hex()}"
            f"&ttl={ttl or self.TTL}"
            f"&topic={request.method.value}"
        )

        data = self.crypto.encrypt(
            request.model_dump(
                by_alias=True,
                exclude_none=True,
            ),
            wallet_app_key.hex(),
        )

        async with aiohttp.ClientSession(headers=self.request_headers) as session:
            async with session.post(
                url,
                data=data,
                ssl=SSL_CONTEXT,
                timeout=timeout,
            ) as response:
                LOG.info(
                    "Sent request %s to the wallet: %s (%s)",
                    request.id,
                    request.method,
                    request,
                )
                return await response.json()

    def parse_message(self, event: MessageEvent) -> BridgeMessage | None:
        """Parse message from the wallet bridge."""

        try:
            wallet_message_event = WalletMessageEvent.model_validate_json(event.data, strict=True)
        except ValidationError:
            LOG.error("Wallet message validation error: %s", event.data)
            return None

        try:
            decoded_data = self.crypto.decrypt(
                wallet_message_event.message, wallet_message_event.sender
            )
        except CryptoError:
            LOG.error("Wallet message data decryption error")
            return None

        try:
            wallet_event = self.type_adapter.validate_json(decoded_data)
        except ValidationError as e:
            LOG.error("Wallet event validation error: %s", e)
            LOG.error("Wallet event data: %s", decoded_data)
            return None

        return BridgeMessage(
            event=wallet_event,
            app_name=self.app_name,
            source=wallet_message_event.sender,
        )

    def handle_event(self, event: MessageEvent) -> None:
        """Handle event."""

        self.last_rpc_event_id = event.last_event_id
        message = self.parse_message(event)

        if message:
            asyncio.create_task(self.queue.put(message))
            LOG.debug("Sent message to the queue: %s", message)

    async def listen(self) -> None:
        """Listen for events from the bridge SSE."""

        LOG.debug("Starting Bridge listening for events: %s", self.app_name)

        iteration_task: asyncio.Task | None = None

        while not self.stop.is_set():
            try:
                url = self.generate_url()
                async with aiohttp.ClientSession() as session:
                    async with aiohttp_sse_client.client.EventSource(
                        url=url,
                        session=session,
                        max_connect_retry=1,
                        ssl=SSL_CONTEXT,
                        reconnection_time=timedelta(seconds=2),
                        on_message=self.handle_event,
                        on_open=self.connected.set,
                    ) as event_source:
                        LOG.info("Bridge connected: %s", self.app_name)

                        await self.connector_ready.wait()
                        iteration_task = asyncio.create_task(iterate_event_source(event_source))

                        while self.connected and not self.stop.is_set():
                            await asyncio.sleep(5)
                            await self.queue.put(
                                BridgeMessage(
                                    event="heartbeat",
                                    app_name=self.app_name,
                                    source=b"",
                                )
                            )

            except asyncio.CancelledError:
                iteration_task.cancel()
                self.stop.set()
                LOG.info("Bridge listener for %s is cancelled.", self.app_name)
                return
            except TimeoutError:
                LOG.error("Bridge connection timeout")
            except Exception as e:
                LOG.exception("Bridge connection unhandled error: %s", e)
            finally:
                if not self.stop.is_set():
                    await self.disconnect(send_event=False)
                    LOG.info("Retrying Bridge connection in 1 second...")
                    await asyncio.sleep(1)

    def generate_url(self) -> str:
        """Generate URL for the bridge."""

        url = f"{self.bridge_url}/{self.PATH_EVENTS}?client_id={self.crypto.public_key}"

        if self.last_rpc_event_id is not None:
            url += f"&last_event_id={self.last_rpc_event_id}"

        return url

    async def register_session(self) -> None:
        """Register session with the bridge."""

        self.listener = asyncio.create_task(self.listen(), name="BridgeListener")



import ssl
from typing import AsyncIterator, TypeVar

import certifi
from aiohttp import ClientConnectionError

R = TypeVar("R")

SSL_CONTEXT = ssl.create_default_context(cafile=certifi.where())


def encode_telegram_url_parameters(parameters: str) -> str:
    """Format bot command url."""

    return (
        parameters.replace(".", "%2E")
        .replace("-", "%2D")
        .replace("_", "%5F")
        .replace("&", "-")
        .replace("=", "__")
        .replace("%", "--")
        .replace("+", "")
    )


async def iterate_event_source(iterator: AsyncIterator) -> None:
    try:
        async for _ in iterator:
            pass
    except ClientConnectionError:
        pass



from typing import Generic, List, Literal, TypeVar, Union

from pydantic import BaseModel, ConfigDict, Field, PositiveInt

from ton_connect.model.misc import GenericLiteral

_FeatureName = TypeVar("_FeatureName")


class Feature(BaseModel, Generic[_FeatureName]):
    model_config = ConfigDict(extra="allow")

    name: GenericLiteral[_FeatureName] = Field(..., description="Name of the feature")


class SendTransactionFeature(Feature[Literal["SendTransaction"]]):
    max_messages: PositiveInt = Field(
        ..., description="Max number of messages", alias="maxMessages"
    )


SignDataFeature = Feature[Literal["SignData"]]

FeatureType = Union[SendTransactionFeature, SignDataFeature, str]


class Device(BaseModel):
    """User's device info."""

    platform: str = Field(..., description="Platform of the user's device")
    app_name: str = Field(..., description="Name of the user's wallet app", alias="appName")
    app_version: str = Field(
        ..., description="Version of the user's wallet app", alias="appVersion"
    )
    max_protocol_version: int = Field(
        ...,
        description="Max protocol version of the user's wallet app",
        alias="maxProtocolVersion",
    )
    features: List[FeatureType] = Field(
        ..., description="List of features of the user's wallet app"
    )



import datetime
from enum import Enum
from typing import Generic, TypeVar

from pydantic import BaseModel, Field

from ton_connect.model.response import ResponseError

_GenericResponseError = TypeVar("_GenericResponseError", bound=ResponseError)
_DataType = TypeVar("_DataType")
_ResultGeneric = TypeVar("_ResultGeneric")


class Action(str, Enum):
    SEND_TRANSACTION = "send_transaction"
    SIGN_DATA = "sign_data"
    DISCONNECT = "disconnect"


class WalletResponseError(BaseModel, Generic[_GenericResponseError]):
    id: str = Field(..., description="Event ID")
    error: _GenericResponseError = Field(..., description="Error message")


class SignDataSuccessResponse(BaseModel):
    signature: str = Field(..., description="Signature of the data")
    timestamp: datetime.datetime = Field(..., description="Timestamp of the signature")
